
(*#######################################*)
 IMPLEMENTATION MODULE RegularExpressions;    (* v1.2-111504 for ISO *)
(*#######################################*)             (* 031297 -> changed SIZE(p^) to SIZE(type) *)
(*
Copyright (C) 1989 Thomas Breeden.

Permission to use, copy, modify and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided that
the above copyright notice appear in all copies and that both the copyright
notice and this permission notice appear in supporting documentation.
Thomas Breeden makes no representations about the suitability of this software
for any purpose. It is provided "as is" without express or implied warranty.
*)
(*
        Support for pattern matching a la AmigaDos.
*)

FROM SYSTEM  IMPORT CAST;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
(*FROM (*System*)ExStorage IMPORT Available;*)
FROM Debugging0 IMPORT Assert;

FROM STextIO  IMPORT WriteChar, WriteString, WriteLn;
FROM SWholeIO IMPORT WriteCard;

FROM Str0     IMPORT StrAsg, StrLen, String256;
FROM Str1     IMPORT ChrCat, StrCap, StrCat, StrClr, StrPos, StrSeg;

IMPORT Set0; (* for StateStack module *)

CONST DEBUG = FALSE;

TYPE  StatePtr = POINTER TO State;
      NFA      = StatePtr;

      StateType = (Start, Final, Interior);

      Edge  = RECORD
                 Label    :CHAR;
                 RangeLo,
                 RangeHi  :CHAR;
                 Dst      :StatePtr;
              END;

      State = RECORD
                 Typ         :StateType;
                 IsSimple    :BOOLEAN;      (* used only if Typ = Start *)
                 Num         :CARDINAL16;
                 Edge1,
                 Edge2       :Edge;
                 FinalState  :StatePtr;      (* used only if Typ = Start *)
                 StackLink   :ARRAY [1..2] OF StatePtr;
              END;

CONST (* Special chars in Regular Expression *)
        AlternateOp     = '|';
      (*ConcatOp        =    -> implied by adjacency *)
        KleeneClosureOp = '#';
        AnyCharOp       = '?';
        QuoteOp         = "'";
        EpsiOp          = '%';
        RangeOp         = '-';
      (* also             '(' or '[' *)
      (* also             ')' or ']' *)
      (* also             '*' *)

        OperatorChars = "|#?%-()[]*";

     (* Special Edge Labels and Chars *)
        RangeSpecial          = 367C;
        OpenParenSpecial      = 370C;
        CloseParenSpecial     = 371C;
        AlternateSpecial      = 372C;
        KleeneClosureSpecial  = 373C;
        AnyCharSpecial        = 374C;
        EndOfStrSpecial       = 375C;
        NoEdgeSpecial         = 376C;
        EpsiSpecial           = 377C;

VAR     StrBuffer :String256;
        Len,
        Pos       :CARDINAL;
        chG       :CHAR;              (*  -->> CAREFUL, chG is a GLOBAL *)
        Loch,
        Hich      :CHAR;

(*
TO DO
(* MAKE StrBuffer handling an embedded MODULE *)
(* TEST FOR MAX BIGSET AND NUMBER OF STATES *)
(* RECOVER MEM IF Err !!!!!!!!!!!!!! *)
(* Different type of stack? *)
(* DOES IT MATTER LEFT ASSOCIATIVITY??? *)
(* NEED A ONE OR MORE OPERATOR *)
(* NEED A NFA VARIABLE OPERATOR *)

NOTE:

   1.This code is a straightforward, non-optimized implementation of algorithms
     3.3 (Thompson's Construction) for producing an NFA from a regular
     expression and 3.4 for simulating an NFA from the new dragon book,
     "Compilers, Principles, Techniques, and Tools", Aho, Sethi, Ullman.

   2.Special properties of the NFA created here are taken advantage of:

         a.It has at most twice as many states as the number of symbols and
           operators in the regular expression.

         b.It has exactly one start state and one accepting state.

         c.Each state has at most two outgoing transition, only one of which
           is non-Episilon.

   3.States are numbered from 1 to <N>. Num 0 is used to mean "not yet
     assigned a number.

     During creation, however, Handle^.Num, the overall start state's Num field,
     is used to contain a count of the total number of states in the NFA.

     Actual emumeration of the states is done at the end of creation. The final
     state always has the Num <N>.

VERSION HISTORY

v1.0-061487 for TDI compiler

v1.0-042389 for Benchmark
                -One INTEGER <-> CARDINAL FOR LOOP CHANGES

v1.1-042389     -Added "range of chars" capability, ie A-Z, 0-9, etc.
                -Now DiscardNFA returns NoErr correctly.

    -111489     -Disabled non-VAR dynamic array copying (*$D-*) for
                 CreateNFA and RunNFA. BE CAREFUL.

v1.2-121300     -For ISO 32bit compiler.

v1.2-121400     -for StonyBrook 4, ExStorage.Available() instead of SystemStorage.Available().
                -Made a "*" and abbreviation for "#?"
                -Now brackets "[" and "]" are treated the same as parentheses.
                -"StrBuffer" for the regexpression is bigger, now String256.
                -Fixed some quote problems by making FillInputBuffer act differently for the RegExpr
                 creation and for the testing.

    -022001     -Added func InvalidNFA().

    -073004   - Added IsSimpleNFA() func, which indicates that no operators were used, and
                may allow saving search time.
              - Changed name of GLOBAL varialble "ch" to "chG"

    -092404     -Replaced system dependent call of Available() with a kludgy system independent
                 call.

    -092804     -Changed OrNFA() to return a BadExpr error if one of the two input NFAs is NIL.
                 ie, don't want to accept things like (A|B|C|).

    -111504     -Avoid doing the code for IsSimple at end of CreateNFA if the NFA
                 after Enumerate comes back NIL.
*)

(*----------------------------------------*)
 PROCEDURE WriteStringLn(s:ARRAY OF CHAR);
(*----------------------------------------*)

BEGIN

WriteString(s);
WriteLn;

END WriteStringLn;

(*--------------------------------------*)
 PROCEDURE Available(n:CARDINAL):BOOLEAN;
(*--------------------------------------*)

VAR  p     :POINTER TO CHAR;
     bRes  :BOOLEAN;

BEGIN

ALLOCATE(p, n);
IF p = NIL THEN
   RETURN FALSE;
END;

DEALLOCATE(p, n);
RETURN TRUE;

END Available;

(*#######################################################################*)

(*#################*)
 MODULE StateStack;
(*#################*)

FROM Set0 IMPORT BigSet, NullBigSet, SetIn, SetIncl;
IMPORT StatePtr;

EXPORT InitStack, Push, Pop, StateIn;

TYPE StackType = RECORD
                   StackTop      :StatePtr;
                   StackedStates :BigSet;
                 END;

VAR Stack :ARRAY[1..2] OF StackType;

(* NOTE: 1.These stacks will stack something only once *)

(*====================================*)
 PROCEDURE InitStack(StackNum:CARDINAL);
(*====================================*)

BEGIN WITH Stack[StackNum] DO

StackTop := NIL;
StackedStates := NullBigSet;

END END InitStack;

(*==============================================*)
 PROCEDURE Push(Sptr:StatePtr; StackNum:CARDINAL);
(*==============================================*)

BEGIN WITH Stack[StackNum] DO

IF SetIn(Sptr^.Num, StackedStates) THEN
   RETURN
END;

Sptr^.StackLink[StackNum] := Stack[StackNum].StackTop;
StackTop := Sptr;

SetIncl(Sptr^.Num, StackedStates);

END END Push;

(*==============================================*)
 PROCEDURE Pop(VAR P:StatePtr; StackNum:CARDINAL);
(*==============================================*)

(* Returns NIL for empty stack *)

BEGIN WITH Stack[StackNum] DO

IF StackTop = NIL THEN
   P := NIL;
ELSE
   P := StackTop;
   StackTop := StackTop^.StackLink[StackNum];
END;

END END Pop;

(*==========================================================*)
 PROCEDURE StateIn(Sptr:StatePtr; StackNum:CARDINAL):BOOLEAN;
(*==========================================================*)

BEGIN WITH Stack[StackNum] DO

RETURN SetIn(Sptr^.Num, StackedStates);

END END StateIn;

(*--------------------------*)
 BEGIN (* MODULE StateStack *)
(*--------------------------*)

InitStack(1); InitStack(2);   (* necessary? *)

END StateStack;

(*#######################################################################*)

(*------------------------------------------------------------*)
 PROCEDURE FillInputBuffer(S:ARRAY OF CHAR; IsRegExpr:BOOLEAN);
(*------------------------------------------------------------*)

VAR  i      :INTEGER;
     quoted :BOOLEAN;
     ch     :CHAR;

BEGIN

StrClr(StrBuffer);
quoted := FALSE;

(* Remove8thBit and substitute some special characters *)
FOR i := 0 TO VAL(INTEGER, LENGTH(S))-1 DO
   ch := S[i];
   IF ch > 177C THEN
      ch := CHR(ORD(ch)-200B);
   END;
   IF IsRegExpr AND NOT quoted THEN
      IF ch = '*' THEN
         StrCat("#?", StrBuffer);
      ELSIF ch = '[' THEN
         ChrCat('(', StrBuffer);
      ELSIF ch = ']' THEN
         ChrCat(')', StrBuffer);
      ELSE
         ChrCat(ch, StrBuffer);
         IF ch = "'" THEN
            quoted := TRUE;
         END;
      END;
   ELSE
      ChrCat(ch, StrBuffer);
      quoted := FALSE;
   END;
END;

Len := LENGTH(StrBuffer);
Pos := 0;

END FillInputBuffer;

(*-----------------------*)
 PROCEDURE NxtChar():CHAR;
(*-----------------------*)

VAR  ch :CHAR;

BEGIN

IF Pos >= Len THEN
   RETURN EndOfStrSpecial;
ELSE
   ch := StrBuffer[Pos];
   INC(Pos);
   RETURN ch;
END;

END NxtChar;

(*--------------------------------*)
 PROCEDURE NxtCharTranslate():CHAR;
(*--------------------------------*)

VAR  ch :CHAR;

BEGIN

IF Pos >= Len THEN
   RETURN EndOfStrSpecial;
ELSE
   ch := StrBuffer[Pos];

   IF ch = AnyCharOp THEN
      ch := AnyCharSpecial;
   ELSIF ch = EpsiOp THEN
      ch := EpsiSpecial;
   ELSIF ch = QuoteOp THEN
      INC(Pos);
      IF Pos < Len THEN
         ch := StrBuffer[Pos];
      END;
   ELSIF ch = AlternateOp THEN
      ch := AlternateSpecial;
   ELSIF ch = KleeneClosureOp THEN
      ch := KleeneClosureSpecial;
   ELSIF ch = '(' THEN
      ch := OpenParenSpecial;
   ELSIF ch = ')' THEN
      ch := CloseParenSpecial;
   END;

   INC(Pos);

   IF (Pos < Len) AND (StrBuffer[Pos] = RangeOp) THEN
      Loch := ch;   (* GLOBALS *)
      Hich := ch;              (* just in case or error *)
      ch := RangeSpecial;
      INC(Pos);
      IF StrBuffer[Pos] = QuoteOp THEN
         INC(Pos);
      END;
      IF Pos < Len THEN
         Hich := StrBuffer[Pos];
         INC(Pos);
      END;
   END;

   RETURN ch;
END;

END NxtCharTranslate;

(*--------------------------------------*)
 PROCEDURE ConstructNFA(E1, E2:CHAR):NFA;
(*--------------------------------------*)

VAR  Result :NFA;

      (*-------------------------------------------------------------------*)
       PROCEDURE ConstructEdge(L:CHAR; VAR StartNum:CARDINAL16; VAR E:Edge);
      (*-------------------------------------------------------------------*)
      BEGIN

      E.Label := L;

      IF L = RangeSpecial THEN         (* assume doing this before next *)
         E.RangeLo := Loch;            (* NxtCharTranslate() call.      *)
         E.RangeHi := Hich;
      END;

      IF L # NoEdgeSpecial THEN
         INC(StartNum);
         ALLOCATE(E.Dst, SIZE(State));
         WITH E.Dst^ DO
            Typ         := Final;
            Num         := 0;
            Edge1.Label := NoEdgeSpecial;
            Edge1.Dst   := NIL;
            Edge2.Label := NoEdgeSpecial;
            Edge2.Dst   := NIL;
            FinalState  := NIL;
            StackLink[1] := NIL;  (* necessary?? *)
            StackLink[2] := NIL;  (* necessary?? *)
         END;
      ELSE
         E.Dst := NIL;
      END;

      END ConstructEdge;

BEGIN

ALLOCATE(Result, SIZE(State));
WITH Result^ DO
   Typ := Start;
   Num := 1;
   ConstructEdge(E1, Num, Edge1);
   ConstructEdge(E2, Num, Edge2);
   FinalState := Edge1.Dst;
   StackLink[1] := NIL;  (* necessary?? *)
   StackLink[2] := NIL;  (* necessary?? *)
END;

RETURN Result;

END ConstructNFA;

(*----------------------------------*)
 PROCEDURE ConcatNFA(N1, N2:NFA):NFA;
(*----------------------------------*)

(* Replace N2 start with N1 final, which becomes interior *)

BEGIN

(* Assert N2^.Typ = Start *)
(* Assert N1^.FinalState^.Edge1/2 = NIL *)

IF N1 = NIL THEN
   RETURN N2
ELSIF N2 = NIL THEN
   RETURN N1
END;

WITH N1^.FinalState^ DO
   Edge1 := N2^.Edge1;
   Edge2 := N2^.Edge2;
   Typ := Interior;
END;
N1^.FinalState := N2^.FinalState;
INC(N1^.Num, N2^.Num-1);

DEALLOCATE(N2, SIZE(State));

RETURN N1;

END ConcatNFA;

(*---------------------------------------------------*)
 PROCEDURE OrNFA(N1, N2:NFA; VAR Err:RegExprErrs):NFA;
(*---------------------------------------------------*)

(* Create new start and final states with Epsi edges to N1 and N2 *)

VAR NewStart, NewFinal :StatePtr;

      (*-----------------------*)
       PROCEDURE Connect(N:NFA);
      (*-----------------------*)
      BEGIN
      WITH N^ DO
         Typ := Interior;
         Num := 0;
         FinalState^.Typ := Interior;
         FinalState^.Edge1.Label := EpsiSpecial;
         FinalState^.Edge1.Dst := NewFinal;
         FinalState := NIL;  (* necessary? *)
      END;
      END Connect;

BEGIN

IF (N1 = NIL) OR (N2 = NIL) THEN
   Err := BadExpr;
   RETURN NIL;
END;

NewStart := CAST(StatePtr, ConstructNFA(NoEdgeSpecial, NoEdgeSpecial));
NewFinal := CAST(StatePtr, ConstructNFA(NoEdgeSpecial, NoEdgeSpecial));

WITH NewStart^ DO
  (*Typ := Start; -> Done by ConstructNFA *)
  Edge1.Label := EpsiSpecial;
  Edge1.Dst := CAST(StatePtr, N1);
  Edge2.Label := EpsiSpecial;
  Edge2.Dst := CAST(StatePtr, N2);
  FinalState := NewFinal;
  INC(Num, N1^.Num + N2^.Num + 1);  (* 1 for NewFinal *)
END;

WITH NewFinal^ DO
   Typ := Final;
   Num := 0;
END;

Connect(N1);
Connect(N2);

RETURN CAST(NFA, NewStart);

END OrNFA;

(*----------------------------------*)
 PROCEDURE KleeneCloseNFA(N:NFA):NFA;
(*----------------------------------*)

(* Create new start and final states with added Epsi edges *)

VAR NewStart, NewFinal :StatePtr;

BEGIN

NewStart := CAST(StatePtr, ConstructNFA(NoEdgeSpecial, NoEdgeSpecial));
NewFinal := CAST(StatePtr, ConstructNFA(NoEdgeSpecial, NoEdgeSpecial));

WITH NewStart^ DO
  (*Typ := Start; -> Done by ConstructNFA *)
  Edge1.Label := EpsiSpecial;
  Edge1.Dst := CAST(StatePtr, N);
  Edge2.Label := EpsiSpecial;
  Edge2.Dst := NewFinal;
  FinalState := NewFinal;
  INC(Num, N^.Num + 1);  (* 1 for NewFinal *)
END;

WITH NewFinal^ DO
   Typ := Final;
   Num := 0;
END;

WITH N^ DO
   Typ := Interior;
   Num := 0;
   WITH FinalState^ DO
      Typ := Interior;
      Edge1.Label := EpsiSpecial;
      Edge1.Dst := NewFinal;
      Edge2.Label := EpsiSpecial;
      Edge2.Dst := CAST(StatePtr, N);
   END;
   FinalState := NIL;
END;

RETURN CAST(NFA, NewStart);

END KleeneCloseNFA;

(*----------------------------*)
 PROCEDURE EnumerateNFA(N:NFA);
(*----------------------------*)


VAR  P      :StatePtr;
     Number :CARDINAL;

      (*-------------------------------*)
       PROCEDURE VisitEdge(Pe:StatePtr);
      (*-------------------------------*)
      BEGIN
      IF (Pe # NIL) AND (Pe^.Num = 0) THEN
         Pe^.Num := Number;
         INC(Number);
         Push(Pe, 1);
      END;
      END VisitEdge;

BEGIN

IF N = NIL THEN
   RETURN;
END;

InitStack(1);

N^.FinalState^.Num := N^.Num;
N^.Num := 1;
Number := 2;

Push(N^.FinalState, 1);
P := CAST(StatePtr, N);
REPEAT
   VisitEdge(P^.Edge1.Dst);
   VisitEdge(P^.Edge2.Dst);
   Pop(P, 1);
UNTIL P = NIL;

END EnumerateNFA;

 PROCEDURE Concatenation(VAR Err:RegExprErrs):NFA;  FORWARD;
 PROCEDURE Alternation(VAR Err:RegExprErrs):NFA;    FORWARD;
 PROCEDURE Closure(VAR Err:RegExprErrs):NFA;        FORWARD;

(*-----------------------------------------------*)
 PROCEDURE Concatenation(VAR Err:RegExprErrs):NFA;
(*-----------------------------------------------*)

VAR   Handle  :NFA;

BEGIN

Handle := Closure(Err);
IF (Err = BadExpr) THEN RETURN NIL END;

IF (chG # AlternateSpecial) AND (chG # CloseParenSpecial)
                           AND (chG # EndOfStrSpecial) THEN
   Handle := ConcatNFA(Handle, Concatenation(Err));
   IF (Err = BadExpr) THEN RETURN NIL END;
END;

RETURN Handle;

END Concatenation;

(*---------------------------------------------*)
 PROCEDURE Alternation(VAR Err:RegExprErrs):NFA;
(*---------------------------------------------*)

VAR   Handle  :NFA;
      Err2    :RegExprErrs;

BEGIN

Handle := Concatenation(Err);
IF (Err = BadExpr) THEN RETURN NIL END;

IF chG = AlternateSpecial THEN
   chG := NxtCharTranslate();
   Err2 := NoErr;
   Handle := OrNFA(Handle, Alternation(Err), Err2);
   IF (Err # NoErr) THEN
      RETURN NIL;
   ELSIF Err2 # NoErr THEN
      Err := Err2;
      RETURN NIL;
   END;
END;

RETURN Handle;

END Alternation;

(*-----------------------------------------*)
 PROCEDURE Closure(VAR Err:RegExprErrs):NFA;
(*-----------------------------------------*)

VAR   Handle  :NFA;

BEGIN

IF chG = KleeneClosureSpecial THEN
   chG := NxtCharTranslate();
   IF chG = OpenParenSpecial THEN
      chG := NxtCharTranslate();
      Handle := KleeneCloseNFA(Alternation(Err));
      IF chG # CloseParenSpecial THEN
         Err := BadExpr;
      ELSE
         chG := NxtCharTranslate();
      END;
   ELSIF (ORD(chG) < 128) OR (chG = AnyCharSpecial) THEN
      Handle := KleeneCloseNFA(ConstructNFA(chG, NoEdgeSpecial));
      chG := NxtCharTranslate();
   ELSE
      Handle := NIL;
      Err := BadExpr;
   END;
ELSIF chG = OpenParenSpecial THEN
   chG := NxtCharTranslate();
   Handle := Alternation(Err);
   IF chG # CloseParenSpecial THEN
      Err := BadExpr;
   ELSE
      chG := NxtCharTranslate();
   END;
ELSIF chG = CloseParenSpecial THEN
   Handle := NIL;
ELSIF chG = EndOfStrSpecial THEN
   Handle := NIL;
ELSE
   Handle := ConstructNFA(chG, NoEdgeSpecial);
   chG := NxtCharTranslate();
END;

RETURN Handle;

END Closure;

(*==========================*)
 PROCEDURE InvalidNFA():NFA;
(*==========================*)

BEGIN

RETURN NIL;

END InvalidNFA;

(*$D-*) (* dynamic array copying *)
(*========================================================*)
 PROCEDURE CreateNFA(VAR Handle:NFA; RegExpr:ARRAY OF CHAR;
                     VAR Err:RegExprErrs);
(*========================================================*)
(*
*)

VAR  chOp  :CHAR;
     i     :CARDINAL;

BEGIN

Handle := NIL;
Err := NoErr;

IF StrLen(RegExpr) = 0 THEN
   Err := BadExpr;
   RETURN;
ELSIF NOT Available(2*StrLen(RegExpr)*SIZE(State)) THEN
   Err := NoMemory;
   RETURN;
END;

FillInputBuffer(RegExpr, TRUE(*IsRegExpr*));

chG := NxtCharTranslate();
Handle := Alternation(Err);

EnumerateNFA(Handle);

IF Handle # NIL THEN

   Assert(Handle^.Typ = Start, "Handle returning not Start?");

   Handle^.IsSimple := TRUE;
   i := 0;
   WHILE Handle^.IsSimple AND (i < LENGTH(OperatorChars)) DO
      StrSeg(OperatorChars, i, i, chOp);
      IF StrPos(RegExpr, chOp, 0) >= 0 THEN
         Handle^.IsSimple := FALSE;
      END;
      INC(i);
   END;

END;

END CreateNFA;
(*$D+*) (* dynamic array copying *)

(*--------------------------------------------------*)
 PROCEDURE Move(From:CARDINAL; ch:CHAR; To:CARDINAL);
(*--------------------------------------------------*)
    (* all states reachable from here given ch *)

VAR P :StatePtr;

         (*--------------------------*)
          PROCEDURE CheckEdge(E:Edge);
         (*--------------------------*)
         BEGIN

         IF (E.Dst # NIL)
            AND
            (  (E.Label = ch)
                   OR
               (E.Label = AnyCharSpecial)
                   OR
               ((E.Label = RangeSpecial)
                         AND (ch >= E.RangeLo) AND (ch <= E.RangeHi))
            ) THEN

            Push(E.Dst, To);

         END;

         END CheckEdge;

BEGIN

InitStack(To);

Pop(P, From);

WHILE P # NIL DO

   CheckEdge(P^.Edge1);
   CheckEdge(P^.Edge2);

   Pop(P, From);
END;

END Move;

(*---------------------------------------*)
 PROCEDURE EpsiClosure(From, To:CARDINAL);
(*---------------------------------------*)

VAR P :StatePtr;

         (*--------------------------*)
          PROCEDURE CheckEdge(E:Edge);
         (*--------------------------*)
         BEGIN

         IF (E.Dst # NIL) AND (E.Label = EpsiSpecial) THEN
            Push(E.Dst, From);  (* for further processing *)
         END;

         END CheckEdge;

BEGIN

InitStack(To);

Pop(P, From);

WHILE P # NIL DO
   Push(P, To);  (* into E-closure *)

   CheckEdge(P^.Edge1);
   CheckEdge(P^.Edge2);

   Pop(P, From);
END;

END EpsiClosure;

(*$D-*) (* dynamic array copying *)
(*===================================================================*)
 PROCEDURE RunNFA(Handle:NFA; Str:ARRAY OF CHAR; VAR Err:RegExprErrs)
                                                              :BOOLEAN;
(*===================================================================*)

VAR  ch  :CHAR;

BEGIN

Err := NoErr;
IF Handle = NIL THEN
   Err := NilNfa;
   RETURN FALSE;
END;

FillInputBuffer(Str, FALSE(*IsRegExpr*));

InitStack(1);

Push(CAST(StatePtr, Handle), 1);
EpsiClosure(1, 2);  (* all states reachable from start via Epsi *)

ch := NxtChar();

WHILE ch # EndOfStrSpecial DO

   Move(2, ch, 1);    (* all states reachable from here given ch *)
   EpsiClosure(1, 2); (* and also via Epsi *)

   ch := NxtChar();
END;

RETURN StateIn(Handle^.FinalState, 2);

END RunNFA;
(*$D+*) (* dynamic array copying *)

(*=========================================*)
 PROCEDURE IsSimpleNFA(Handle:NFA):BOOLEAN;
(*=========================================*)

BEGIN

IF Handle = NIL THEN
   RETURN FALSE;
END;

RETURN Handle^.IsSimple;

END IsSimpleNFA;

(*========================================================*)
 PROCEDURE DiscardNFA(VAR Handle:NFA; VAR Err:RegExprErrs);
(*========================================================*)

VAR  P      :StatePtr;

      (*-------------------------------*)
       PROCEDURE VisitEdge(Pe:StatePtr);
      (*-------------------------------*)
      BEGIN
      IF (Pe # NIL) THEN
         Push(Pe, 1);
         Push(Pe, 2);
      END;
      END VisitEdge;

BEGIN

Err := NoErr;
IF Handle = NIL THEN
   Err := NilNfa;
   RETURN
END;

InitStack(1);
InitStack(2);

Push(CAST(StatePtr, Handle), 1);
Push(CAST(StatePtr, Handle), 2);
P := CAST(StatePtr, Handle);
REPEAT
   VisitEdge(P^.Edge1.Dst);
   VisitEdge(P^.Edge2.Dst);
   Pop(P, 1);
UNTIL P = NIL;

Pop(P, 2);
WHILE P # NIL DO
   DEALLOCATE(P, SIZE(State));
   Pop(P, 2);
END;

END DiscardNFA;

(*=====================================*)
(*DEBUG*) PROCEDURE DumpNFA(Handle:NFA);
(*=====================================*)

VAR  P   :StatePtr;

      (*--------------------------*)
       PROCEDURE WriteEdge(E:Edge);
      (*--------------------------*)
      BEGIN WITH E DO
      IF Dst = NIL THEN
         RETURN
      END;
      WriteString('Edge to '); WriteCard(Dst^.Num, 0);
      WriteString('   labeled: ');
      IF Label = EpsiSpecial THEN
         WriteStringLn('Epsi');
      ELSIF Label = AnyCharSpecial THEN
         WriteStringLn('AnyChar');
      ELSIF Label = RangeSpecial THEN
         WriteString('Range ');
         WriteChar(RangeLo); WriteChar('-'); WriteChar(RangeHi); WriteLn;
      ELSIF Label > ' ' THEN
         WriteChar(Label); WriteLn;
      ELSE
         WriteString('CHR('); WriteCard(ORD(Label),0); WriteChar(')'); WriteLn;
      END;
      Push(Dst, 1);
      END END WriteEdge;

BEGIN

InitStack(1);
P := CAST(StatePtr, Handle);

WriteLn;

WHILE P # NIL DO
   WITH P^ DO
      WriteString('STATE #'); WriteCard(Num, 0); WriteLn;
      WriteString('Typ = ');
      CASE Typ OF
         Start:    WriteStringLn('Start');
       | Final:    WriteStringLn('Final');
       | Interior: WriteStringLn('Interior');
      END;
      WriteEdge(Edge1);
      WriteEdge(Edge2);
      WriteLn;
   END;

   Pop(P, 1);
END;

END DumpNFA;

END RegularExpressions.

